#!/usr/bin/env python

import numpy as np
import os
from tqdm import tqdm
import csv
from utils import Color, denoise, iterativeMean, isARawImage, escapeFilePath, getColorChannel, saveNpArray, rescaleRawImageForDenoiser, updateExtremes
import matplotlib.pyplot as plt
from scipy.ndimage import gaussian_filter

imagesFolderPath = 'rafael/arw'
imagesFolderPathFileName = escapeFilePath(imagesFolderPath)
# Among:
# `denoise` possible denoisers and `mean`.
denoiser = 'mean'

raiseNotFlatFields = False
# `[Color.RED, Color.GREEN_RIGHT, ...]` or `Color` or `[None]` for not raw images.
colors = [None]

imagesFileNames = os.listdir(imagesFolderPath + ('/png' if raiseNotFlatFields else ''))

if raiseNotFlatFields:
    files = {}

    with open('RAISE_all.csv') as csvfile:
        reader = csv.DictReader(csvfile)
        for row in tqdm(list(reader), 'CSV parsing'):
            file = row['File'] + '.png'
            files[file] = row

    imagesFileNames = [imageFileName for imageFileName in tqdm(imagesFileNames, 'Filtering images') if files[imageFileName]['Device'] == 'Nikon D7000' and Image.open(f'{imagesFolderPath}/png/{imageFileName}').size == (4946, 3278)]

# Among:
# - `None`
# - `'sky'`
# - `'wall'`
type_ = None
if type_ is not None:
    ranges = {
        'sky': range(2_699, 2_807),
        'wall': range(2_807, 2_912),
    }
    imagesFileNames = [f'DSC0{imageIndex}.ARW' for imageIndex in ranges[type_]]
    imagesFolderPathFileName += f'_{type_}'

minColor = None
maxColor = None

def getImageNpArray(imageFileName, computeExtremes, color):
    global minColor, maxColor
    if raiseNotFlatFields:
        imageFileName = imageFileName.replace('.png', '.NEF')
        imageFilePath = f'{imagesFolderPath}/nef/{imageFileName}'
    else:
        imageFilePath = f'{imagesFolderPath}/{imageFileName}'
    imageNpArray = getColorChannel(imageFilePath)

    if computeExtremes:
        minColor, maxColor = updateExtremes(imageNpArray, minColor, maxColor)
        return

    if isARawImage(imageFileName) and denoiser != 'mean':
        imageNpArray = rescaleRawImageForDenoiser(imageNpArray, minColor, maxColor)
        # Pay attention to range of values expected by the denoiser.
        # Indeed if provide the thousands valued raw image, then the denoiser only returns values between 0 and 1 and making the difference between both look pointless.
    return imageNpArray

# `color` is the actual color to estimate PRNU with.
def treatImage(imageFileName, computeExtremes = False, color = None):
    global estimatedPrnuIterativeMean
    imageNpArray = getImageNpArray(imageFileName, computeExtremes, color)
    if imageNpArray is None:
        return
    if denoiser != 'mean':
        imageDenoisedNpArray = denoise(imageNpArray, denoiser)
    else:
        imageDenoisedNpArray = means[color]
    imageNoiseNpArray = imageNpArray - imageDenoisedNpArray
    estimatedPrnuIterativeMean.add(imageNoiseNpArray)

if (minColor is None or maxColor is None) and denoiser != 'mean':
    # Assuming same intensity scale across color channels.
    for imageFileName in tqdm(imagesFileNames, 'Computing extremes of images'):
        for color in colors:
            treatImage(imageFileName, computeExtremes = True, color = color)

    # To skip this step next time.
    # Maybe thanks to `rawpy.RawPy` fields, possibly stating device maximal value, can avoid doing so to some extent.
    print(f'{minColor=}')
    print(f'{maxColor=}')

if denoiser == 'mean':
    means = {}
    for color in colors:
        colorIterativeMean = iterativeMean()
        for imageFileName in tqdm(imagesFileNames, f'Computing mean of {color} colored images'):
            imageNpArray = getImageNpArray(imageFileName, False, color)
            imageNpArray = gaussian_filter(imageNpArray, sigma = 5)
            colorIterativeMean.add(imageNpArray)
        means[color] = colorIterativeMean.mean
        fileName = f'mean_{imagesFolderPathFileName}_{color}'
        # Then use `merge_single_color_channel_images_according_to_bayer_filter.py` to consider all color channels, instead of saving this single color channel as an image.
        saveNpArray(fileName, colorIterativeMean.mean)

for color in colors:
    estimatedPrnuIterativeMean = iterativeMean()

    for imageFileName in tqdm(imagesFileNames, f'Denoising images for color {color}'):
        treatImage(imageFileName, color = color)

    npArrayFilePath = f'mean_{imagesFolderPathFileName}_{denoiser}_{color}'
    saveNpArray(npArrayFilePath, estimatedPrnuIterativeMean.mean)
